#!/usr/bin/env python
# gpaw-driver facilitates computations with GPAW
# Copyright (C) 2015 Toon Verstraelen
#
# gpaw-driver is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# gpaw-driver is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>


def main():
    args = parse_args()

    # Look for previous output to be used as restart
    import os
    if os.path.isfile('restart.gpw'):
        filename = 'restart.gpw'
    else:
        filename = None

    # configure k-points
    if args.kpts is None:
        kpts = None
    else:
        kpts = {'size': [int(w) for w in args.kpts.split('.')]}

    import ase.io
    # load atoms
    atoms = ase.io.read(args.geo_in)
    if args.geo_in.endswith('.xyz'):
        load_cell_from_xyz(args.geo_in, atoms, args.margin)

    # Set the initial magnetic moments
    if args.magmom is not None:
        configure_initial_magnetic_moments(atoms, args.magmom)

    from gpaw import GPAW
    # create the GPAW calculator
    kwargs = {
        'kpts': kpts,
        'xc': args.xc,
        'convergence': {
            'energy': args.conv_energy,
            'density': args.conv_density,
            'eigenstates': args.conv_eigenstates,
        },
    }
    if args.mode == 'pw':
        from gpaw import PW
        kwargs['mode'] = PW(args.cutoff)
        # Currently, rmm-diis is the only one that supports band parallelization
        kwargs['eigensolver'] = 'rmm-diis'
        # A Gaussian basis set is needed for the initial guess of the plane
        # waves. The default basis set, i.e. the partial waves, is often not
        # large enough for this purpose. Hence dzp is used.
        kwargs['basis'] = 'dzp'
    elif args.mode == 'rg':
        from gpaw.utilities import h2gpts
        from gpaw.eigensolvers import Davidson
        # Manually set number of grid points, mostly relevant for corner cases
        kwargs['gpts'] = h2gpts(args.spacing, atoms.get_cell(), idiv=args.h2gpts)
        # A more accurate eigensolver than the default.
        kwargs['eigensolver'] = Davidson(3)
    else:
        raise NotImplementedError
    calc = GPAW(filename, **kwargs)
    # attach restart writer to calculator
    calc.attach(calc.write, 1, 'output.gpw')
    # tell the atoms what their PES will be
    atoms.set_calculator(calc)
    if args.runtype == 'energy':
        # compute energy
        atoms.get_potential_energy()
        print ('The atomic forces:')
        print (atoms.get_forces())
    elif args.runtype == 'opt':
        from ase.optimize import BFGS
        opt = BFGS(atoms, trajectory='opt.traj', logfile='opt.log')
        opt.run(fmax=args.opt_fmax)
        atoms.write('opt_final.xyz')
        atoms.write('opt_final.pdb')
        atoms.write('opt_final.cif')
    elif args.runtype == 'cellopt':
        # optimize the structure (atoms and cell)
        # NOT TESTED YET!!
        from ase.optimize import BFGS
        from ase.constraints import UnitCellFilter
        from ase.io.trajectory import Trajectory
        ucf = UnitCellFilter(atoms)
        traj = Trajectory('cellopt.traj', 'w', atoms)
        opt = BFGS(ucf)
        opt.attach(traj)
        opt.run(fmax=args.opt_fmax)
        atoms.write('opt_final.xyz')
        atoms.write('opt_final.pdb')
        atoms.write('opt_final.cif')
    else:
        raise NotImplementedError
    if args.store_all:
        calc.write('output.gpw', mode='all')


def parse_args():
    import argparse
    parser = argparse.ArgumentParser(prog='gpaw-driver',
        description='Run a GPAW single-point calculation.')
    parser.add_argument('geo_in',
        help='A geometry input file. If XYZ, cell vecs will be read from title')

    parser.add_argument('--dry-run', default=None, type=int,
        help='When given, the computation is tested as if DRY_RUN processes '
             'were used.')

    parser.add_argument('--mode', default='rg', choices=['rg', 'pw'],
        help='GPAW Mode: real-space grid (rg) or plane waves (pw). [default=%(default)s]')
    parser.add_argument('--spacing', default=0.2, type=float,
        help='The real-space grid spacing in angstrom (only relevant in rg '
             'mode). [default=%(default)s]')
    parser.add_argument('--h2gpts', default=16, type=int,
        help='The number of grid points will be a multiple of this argument '
             '(only relevant in rg mode). [default=%(default)s].')
    parser.add_argument('--cutoff', default=400, type=float,
        help='The plane-wave cutoff in electronvolt (only relevant in pw '
             'mode). [default=%(default)s].')

    parser.add_argument('--xc', default='PBE',
        help='The XC functional. [default=%(default)s]')
    parser.add_argument('--margin', default=4.0, type=float,
        help='Vacuum margin for aperiodic dimensions. [default=%(default)s]')
    parser.add_argument('--magmom', default=None, type=str, nargs='+',
        help='Set initial magnetic moments for some elements. Format: "atom=magmom", can '
             'be repeated. When atom is an element symbol, all atoms of that type will '
             'get the same initial value. When atom is an integer, it refers to the atom '
             'index.')
    parser.add_argument('--conv-energy', default=0.0005, type=float,
        help='The energy convergence in eV per electron. [default=%(default)s]')
    parser.add_argument('--conv-density', default=0.0001, type=float,
        help='The density convergence in electron per valence electron. '
             '[default=%(default)s]')
    parser.add_argument('--conv-eigenstates', default=4.0e-8, type=float,
        help='The eigenstat convergence in eV^2 per electron. [default=%(default)s]')
    parser.add_argument('--kpts', default=None,
        help='The Monkhorst-pack k-point grid to be used. If not given, gamma-point '
             'sampling is used. Format: "n1.n2.n3".')

    parser.add_argument('--runtype', default='energy', choices=['energy', 'opt', 'cellopt'],
        help='Run type: energy (single point calculation of energy and forces), opt '
             '(optimization of nuclear coordinates), cellopt (optimization of cell '
             'parameters and nuclear coordinates). [default=%(default)s]')
    parser.add_argument('--opt-fmax', default=0.025, type=float,
        help='The maximum residual force acting on a Cartesian coordinate after '
             'a structure optimization. [default=%(default)s]')
    parser.add_argument('--store-all', default=False, action='store_true',
        help='Store all results in the gpw output, including bands.')
    return parser.parse_args()


def load_cell_from_xyz(fn, atoms, margin):
    import numpy as np
    import shlex

    with open(fn) as f:
        # Load data from XYZ title line into title_info dictionary
        f.next()  # skip number of atoms
        title_info = {}
        for word in shlex.split(f.next()):
            key, _, value = word.rpartition('=')
            title_info[key] = value

        # Convert some data in the title_info dictionary in more useful python types
        if 'Lattice' in title_info:
            vecs = [float(word) for word in title_info['Lattice'].split()]
            vecs = list(np.array(vecs).reshape(-1, 3))
        else:
            vecs = []
        if 'pbc' in title_info:
            pbc = [bool(int(word.lower().replace('t', '1').replace('f', '0')))
                   for word in title_info['pbc'].split()]
        else:
            pbc = [True]*len(vecs)

        # Add cell vectors where needed to obtain a box
        if len(vecs) == 0:
            for ivec in xrange(3):
                basis = np.zeros(3, float)
                basis[ivec] = 1.0
                proj = np.dot(atoms.positions, basis)
                vec = basis*(proj.max() - proj.min() + 2*margin)
                vecs.append(vec)
                pbc.append(False)
                offset = margin - proj.min()
                atoms.positions += offset*basis
        elif len(vecs) == 1:
            # Get two new cell vectors orthogonal to the given one
            imin = abs(vecs[0]).argmin()
            tmp = np.zeros(3)
            tmp[imin] = 1.0
            basis2 = np.cross(vecs[0], tmp)
            basis2 /= np.linalg.norm(basis2)
            basis1 = np.cross(basis2, vecs[0])
            basis1 /= np.linalg.norm(basis1)
            # Adjust the length of the cell vectors to enclose the system with
            # a given margin.
            proj1 = np.dot(atoms.positions, basis1)
            vec1 = basis1*(proj1.max() - proj1.min() + 2*margin)
            vecs.append(vec1)
            pbc.append(False)
            proj2 = np.dot(atoms.positions, basis2)
            vec2 = basis2*(proj2.max() - proj2.min() + 2*margin)
            vecs.append(vec2)
            pbc.append(False)
            # Translate the atoms to fit in the center of the box
            offset1 = margin - proj1.min()
            offset2 = margin - proj2.min()
            atoms.positions += offset1*basis1
            atoms.positions += offset2*basis2
        elif len(vecs) == 2:
            basis2 = np.cross(vecs[0], vecs[1])
            basis2 /= np.linalg.norm(basis2)
            # Adjust the length of the cell vectors to enclose the system with
            # a given margin.
            proj2 = np.dot(atoms.positions, basis2)
            vec2 = basis2*(proj2.max() - proj2.min() + 2*margin)
            vecs.append(vec2)
            pbc.append(False)
            # Translate the atoms to fit in the center of the box
            offset2 = margin - proj2.min()
            atoms.positions += offset2*basis2
        elif len(vecs) == 3:
            pass
        else:
            raise NotImplementedError

        print (vecs)
        atoms.set_cell(np.array(vecs))
        atoms.set_pbc(np.array(pbc))


def configure_initial_magnetic_moments(atoms, magmom_settings):
    from ase.data import atomic_numbers
    import string
    for setting in magmom_settings:
        if setting.count('=') != 1:
            raise ValueError('Magmom settings must contain exactly one =.')
        atom, magmom = setting.split('=')
        magmom = float(magmom)
        number = atomic_numbers.get(string.capwords(atom))
        if number is None:
            indexes = [int(atom)]
        else:
            indexes = (atoms.numbers == number).nonzero()[0]
        for index in indexes:
            atoms[index].magmom = magmom


if __name__ == '__main__':
    main()

